// Squawk messenger. 
// Copyright (C) 2019  Yury Gubich <blue@macaw.me>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CORE_CACHE_HPP
#define CORE_CACHE_HPP
#include "cache.h"

template <class K, class V>
Core::Cache<K, V>::Cache(const QString& name):
    storage(name),
    cache(new std::map<K, V> ()),
    abscent(new std::set<K> ()) {}

template <class K, class V>
Core::Cache<K, V>::~Cache() {
    close();
    delete cache;
    delete abscent;
}

template <class K, class V>
void Core::Cache<K, V>::open() {
    storage.open();}

template <class K, class V>
void Core::Cache<K, V>::close() {
    storage.close();}

template <class K, class V>
void Core::Cache<K, V>::addRecord(const K& key, const V& value) {
    storage.addRecord(key, value);
    cache->insert(std::make_pair(key, value));
    abscent->erase(key);
}

template <class K, class V>
V Core::Cache<K, V>::getRecord(const K& key) const {
    typename std::map<K, V>::const_iterator itr = cache->find(key);
    if (itr == cache->end()) {
        if (abscent->count(key) > 0) {
            throw Archive::NotFound(std::to_string(key), storage.getName().toStdString());
        }

        try {
            V value = storage.getRecord(key);
            itr = cache->insert(std::make_pair(key, value)).first;
        } catch (const Archive::NotFound& error) {
            abscent->insert(key);
            throw error;
        }
    }

    return itr->second;
}

template<class K, class V>
bool Core::Cache<K, V>::checkRecord(const K& key) const {
    typename std::map<K, V>::const_iterator itr = cache->find(key);
    if (itr != cache->end())
        return true;

    if (abscent->count(key) > 0)
        return false;

    try {
        V value = storage.getRecord(key);
        itr = cache->insert(std::make_pair(key, value)).first;
    } catch (const Archive::NotFound& error) {
        return false;
    }

    return true;
}

template<class K, class V>
void Core::Cache<K, V>::changeRecord(const K& key, const V& value) {
    storage.changeRecord(key, value);   //there is a non straightforward behaviour: if there was no element at the sorage it will be added
    cache->at(key) = value;
    abscent->erase(key);                //so... this line here is to make it coherent with the storage
}

template<class K, class V>
void Core::Cache<K, V>::removeRecord(const K& key) {
    storage.removeRecord(key);
    cache->erase(key);
    abscent->insert(key);
}

#endif //CORE_CACHE_HPP
